# -*- coding: utf-8 -*-
"""
/***************************************************************************
 BioDispersal
                                 A QGIS plugin
 Computes ecological continuities based on environments permeability
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                              -------------------
        begin                : 2018-04-12
        git sha              : $Format:%H$
        copyright            : (C) 2018 by IRSTEA
        email                : mathieu.chailloux@irstea.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

import os.path
import sys
import inspect

from PyQt5.QtCore import QSettings, QTranslator, qVersion, QCoreApplication
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QAction

# Initialize Qt resources from file resources.py
from .resources import *
# Import the code for the dialog
from .BioDispersal_dialog import BioDispersalDialog
from .algs.BioDispersal_algs_provider import BioDispersalAlgorithmsProvider
from .qgis_lib_mc import utils

from qgis.utils import qgis_excepthook
from qgis.core import QgsApplication


cmd_folder = os.path.split(inspect.getfile(inspect.currentframe()))[0]

if cmd_folder not in sys.path:
    sys.path.insert(0, cmd_folder)

class BioDispersal:
    """QGIS Plugin Implementation."""

    provider = None
    
    def __init__(self, iface):
        """Constructor.

        :param iface: An interface instance that will be passed to this class
            which provides the hook by which you can manipulate the QGIS
            application at run time.
        :type iface: QgsInterface
        """
        # Save reference to the QGIS interface
        self.iface = iface
        # initialize plugin directory
        self.plugin_dir = os.path.dirname(__file__)
        # initialize locale
        locale = QSettings().value('locale/userLocale')[0:2]
        utils.info("locale = " + str(locale))
        locale_path = os.path.join(
            self.plugin_dir,
            'i18n',
            'BioDispersal_{}.qm'.format(locale))
        utils.info("locale_path = " + str(locale_path))

        if os.path.exists(locale_path):
            self.translator = QTranslator()
            self.translator.load(locale_path)

            if qVersion() > '4.3.3':
                QCoreApplication.installTranslator(self.translator)

        self.initProcessing()

        # Create the dialog (after translation) and keep reference
        # self.dlg = BioDispersalDialog()
        utils.debug("init")
        #self.dlg = BioDispersalDialog()
        #self.dlg = TestDialog()

        # Declare instance attributes
        self.actions = []
        self.pluginName = self.tr(u'&BioDispersal')
        self.menu = self.pluginName
        # TODO: We are going to let the user set this up in a future iteration
        self.toolbar = self.iface.addToolBar(u'BioDispersal')
        self.toolbar.setObjectName(u'BioDispersal')

    # noinspection PyMethodMayBeStatic
    def tr(self, message):
        """Get the translation for a string using Qt translation API.

        We implement this ourselves since we do not inherit QObject.

        :param message: String for translation.
        :type message: str, QString

        :returns: Translated version of message.
        :rtype: QString
        """
        # noinspection PyTypeChecker,PyArgumentList,PyCallByClass
        return QCoreApplication.translate('BioDispersal', message)


    def add_action(
        self,
        icon_path,
        text,
        callback,
        enabled_flag=True,
        add_to_menu=True,
        add_to_toolbar=True,
        status_tip=None,
        whats_this=None,
        parent=None):
        """Add a toolbar icon to the toolbar.

        :param icon_path: Path to the icon for this action. Can be a resource
            path (e.g. ':/plugins/foo/bar.png') or a normal file system path.
        :type icon_path: str

        :param text: Text that should be shown in menu items for this action.
        :type text: str

        :param callback: Function to be called when the action is triggered.
        :type callback: function

        :param enabled_flag: A flag indicating if the action should be enabled
            by default. Defaults to True.
        :type enabled_flag: bool

        :param add_to_menu: Flag indicating whether the action should also
            be added to the menu. Defaults to True.
        :type add_to_menu: bool

        :param add_to_toolbar: Flag indicating whether the action should also
            be added to the toolbar. Defaults to True.
        :type add_to_toolbar: bool

        :param status_tip: Optional text to show in a popup when mouse pointer
            hovers over the action.
        :type status_tip: str

        :param parent: Parent widget for the new action. Defaults None.
        :type parent: QWidget

        :param whats_this: Optional text to show in the status bar when the
            mouse pointer hovers over the action.

        :returns: The action that was created. Note that the action is also
            added to self.actions list.
        :rtype: QAction
        """

        icon = QIcon(icon_path)
        action = QAction(icon, text, parent)
        action.triggered.connect(callback)
        action.setEnabled(enabled_flag)

        if status_tip is not None:
            action.setStatusTip(status_tip)

        if whats_this is not None:
            action.setWhatsThis(whats_this)

        if add_to_toolbar:
            self.toolbar.addAction(action)

        if add_to_menu:
            self.iface.addPluginToMenu(
                self.menu,
                action)

        self.actions.append(action)

        return action

    def initProcessing(self):
        self.provider = BioDispersalAlgorithmsProvider()
        QgsApplication.processingRegistry().addProvider(self.provider)
    
    def initGui(self):
        # utils.debug("initGui")
        """Create the menu entries and toolbar icons inside the QGIS GUI."""

        icon_path = ':/plugins/BioDispersal/icons/cerf.png'
        # icon_path = os.path.join(os.path.join(cmd_folder, 'cerf.png'))
        self.add_action(
            icon_path,
            text=self.pluginName,
            callback=self.run,
            parent=self.iface.mainWindow())
            
        self.iface.mainWindow()
            
        self.connectComponents()
        
    def initDlgGui(self,dlg):
        dlg.initTabs()
        dlg.initLog()
        dlg.initGui()
        locale = QSettings().value('locale/userLocale')[0:2]
        if locale.startswith('fr'):
            dlg.switchLangFr()
        else:
            dlg.switchLangEn()
        dlg.connectComponents()


    def unload(self):
        #for tab in self.dlg.tabs:
        #    tab.unload()
        """Removes the plugin menu item and icon from QGIS GUI."""
        for action in self.actions:
            self.iface.removePluginMenu(
                self.tr(u'&BioDispersal'),
                action)
            self.iface.removeToolBarIcon(action)
        # remove provider
        if self.provider:
            QgsApplication.processingRegistry().removeProvider(self.provider)
        # remove the toolbar
        utils.print_func = print
        sys.excepthook = qgis_excepthook
        #if self.dlg:
        #    utils.info("unload")
        #    self.dlg.groupsConnector.disconnectComponents()
        #    self.dlg.stConnector.disconnectComponents()
        del self.toolbar

    def connectComponents(self):
        pass
        # self.dlg.connectComponents()
                
    def run(self):
        utils.debug("run")
        """Run method that performs all the real work"""
        # show the dialog
        dlg = BioDispersalDialog()
        self.initDlgGui(dlg)
        dlg.show()
        print(str(dlg))
        # Run the dialog event loop
        result = dlg.exec_()
        # See if OK was pressed
        if result:
            # Do something useful here - delete the line containing pass and
            # substitute with your code.
            pass
            
